/*
Akira Tanaka <akr@jaist.ac.jp> created for gs5.10-vflib-1.4. 1998/05/13
*/

#include "stdio_.h"
#include "memory_.h"
#include "gsmemory.h"
#include "gp.h"
#include "stream.h"
#include "iref.h"
#include "files.h"
#include "gx.h"
#include "gserrors.h"
#include "gxdevice.h"
#include "gxxfont.h"
#include "gxxfvf.h"
#include "kfutil.h"

private xfont_proc_lookup_font(vf_lookup_font);
private xfont_proc_char_xglyph(vf_char_xglyph);
private xfont_proc_char_metrics(vf_char_metrics);
private xfont_proc_render_char(vf_render_char);
private xfont_proc_release(vf_release);
private gx_xfont_procs vflib_xfont_procs = {
        vf_lookup_font,
        vf_char_xglyph,
        vf_char_metrics,
        vf_render_char,
        vf_release,
        NULL,  /* AAAAA */
};

typedef struct vflib_xfont_s vflib_xfont;
struct vflib_xfont_s {
  gx_xfont_common common;
  gs_matrix mat;
  int font_fd;
  int high;
  int wmode;
  int sjis;
};
gs_private_st_simple(st_vflib_xfont, vflib_xfont, "vflib_xfont");

/* table for converting hexadecimal char to number */
#define NOXROW -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
static int x2d[256] = {
  NOXROW NOXROW NOXROW
   0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1, -1, -1, -1, -1, -1,
  -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  NOXROW NOXROW NOXROW
  NOXROW NOXROW NOXROW NOXROW NOXROW NOXROW NOXROW NOXROW
};
#undef NOXROW

gx_xfont *
vf_lookup_font_by_name(
  const byte *fname,
  uint len,
  int encoding_index,
  const gs_uid *puid,
  const gs_matrix *pmat,
  gs_memory_t *mem)
{
  vflib_xfont *vxf;
  int font_fd;
  int high;
  int wmode;
  int sjis;

  const byte *p;
  char *name;
  int nlen;

  /* Is fname the form "*.r[0-9A-F][0-9A-F] or *.r[0-9A-F][0-9A-F]v" ? */
  /*
    Acceptable font name:
      *.rHH       -> JIS or EUC HORIZONTAL
      *.rHHv      -> JIS or EUC VERTICAL
      *.srHH      -> SJIS       HORIZONTAL
      *.srHHv     -> SJIS       VERTICAL

      HH is [0-9A-F][0-9A-F]
  */

  p = fname + len;
  if(p <= fname + 1) /* There is no char. */
    return NULL;
  wmode = p[-1] == 'v' ? (p--, 1) : 0;

  if(p <= fname + 2 && (x2d[p[-1]] < 0 || x2d[p[-2]] < 0))
    return NULL;
  high = (x2d[p[-2]] << 4) + x2d[p[-1]];
  p -= 2;

  if(p <= fname + 1 && p[-1] != 'r')
    return NULL;
  p--;

  if(p <= fname + 1)
    return NULL;
  sjis = p[-1] == 's' ? (p--, 1) : 0;

  if(p <= fname + 1 && p[-1] != '.')
    return NULL;
  p--; /* p points '.' */

  if(sjis)
  {
    if(!(0x81 <= high && high <= 0x9f) &&
       !(0xe0 <= high && high <= 0xfc))
      return NULL;
  }
  else
  {
    if(!(0x21 <= high && high <= 0x7e))
      return NULL;
  }
  
  /* Is pmat simple? */
  if(!(is_xxyy(pmat) || is_xyyx(pmat)))
    return NULL;

  /* Does VFlib provide the font "GS-*" ? */
  nlen = sizeof("GS-") - 1 + (p - fname);
  if((name = (char *)malloc(nlen + 1)) == NULL)
    return NULL;

  memcpy(name, "GS-", sizeof("GS-") - 1);
  memcpy(name + sizeof("GS-") - 1, fname, p - fname);
  name[nlen] = '\0';

  font_fd = vf_open_font(name, nlen, 0);
  if(font_fd < 0)
    return NULL;

  free((void *)name);

  vxf = gs_alloc_struct(mem, vflib_xfont, &st_vflib_xfont, "vflib_lookup_font");
  if(vxf == 0)
    return NULL;
  vxf->common.procs = &vflib_xfont_procs;
  vxf->mat = *pmat;
  vxf->font_fd = font_fd;
  vxf->high = high;
  vxf->wmode = wmode;
  vxf->sjis = sjis;
  return (gx_xfont *)vxf;
}

/* dummy lookup method */
private gx_xfont *
vf_lookup_font(
  gx_device *dev,
  const byte *fname,
  uint len,
  int encoding_index,
  const gs_uid *puid,
  const gs_matrix *pmat,
  gs_memory_t *mem)
{
  return vf_lookup_font_by_name(fname, len, encoding_index, puid, pmat, mem);
}

#define CHAR_GLYPH(jis, wmode) ((jis) | (wmode) << 15)
#define GLYPH_JIS(glyph) ((glyph) & 0x7f7f)
#define GLYPH_WMODE(glyph) (((glyph) & 0x8000) >> 15)

private gx_xglyph
vf_char_xglyph(
  gx_xfont *xf,
  gs_char low,
  int encoding_index,
  gs_glyph glyph,
  gs_proc_glyph_name_t glyph_name_proc)
{
  vflib_xfont *vxf = (vflib_xfont *)xf;
  int jis_code;
  int wmode;

  if(low == gs_no_char)
    return gx_no_xglyph;

  if(vxf->sjis)
  {
    int s1 = vxf->high;
    int s2 = low;
    int j1, j2;

    if(!(0x40 <= low && low <= 0x7e) &&
       !(0x80 <= low && low <= 0xfc))
      return gx_no_xglyph;

    if(s2 < 0x9f)
    {
      j1 = 0x21 - 0x81 * 2 + 0;
      j2 = 0x21 - 0x40 + 0;
    }
    else
    {
      j1 = 0x21 - 0x81 * 2 + 1;
      j2 = 0x21 - 0x40 - 94;
    }
    if(0xe0 <= s1) s1 -= 0xe0 - 0x9f - 1;
    if(0x80 <= s2) s2 -= 0x80 - 0x7e - 1;
    j1 += s1 * 2;
    j2 += s2;
    jis_code = (j1 << 8) | j2;
  }
  else
  {
    if(!(0x21 <= low && low <= 0x7e) &&
       !(0xa1 <= low && low <= 0xfe))
      return gx_no_xglyph;

    low &= 0x7f;
    jis_code = (vxf->high << 8) | low ;
  }

  wmode = vxf->wmode && kf_is_vchar(jis_code);
  return CHAR_GLYPH(jis_code, wmode);
}

private int
vf_char_metrics(
  gx_xfont *xf,
  gx_xglyph xg,
  int paint_wmode,
  gs_point *pwidth,
  gs_int_rect *pbbox)
{
  vflib_xfont *vxf = (vflib_xfont *)xf;
  byte *buf;
  int raster;

  return vf_paint_font(
    vxf->font_fd,
    &vxf->mat,
    GLYPH_JIS(xg),
    GLYPH_WMODE(xg),
    paint_wmode,
    &buf,
    &raster,
    pbbox,
    pwidth);
}

private int
vf_render_char(
  gx_xfont *xf,
  gx_xglyph xg,
  int wmode,
  gx_device *dev,
  int xo,
  int yo,
  gx_color_index color,
  int required)
{
  vflib_xfont *vxf = (vflib_xfont *)xf;
  byte *buf;
  int raster;
  int code;
  gs_int_rect bbox;
  dev_proc_copy_mono((*copy_mono)) = dev_proc(dev, copy_mono);

  if(!required)
    return -1;

  code = vf_paint_font(
    vxf->font_fd,
    &vxf->mat,
    GLYPH_JIS(xg),
    GLYPH_WMODE(xg),
    wmode,
    &buf,
    &raster,
    &bbox,
    NULL);

  if(code < 0)
    return code;

  code = (*copy_mono)(dev, buf, 0, raster, gx_no_bitmap_id,
    xo + bbox.p.x, yo + bbox.p.y, bbox.q.x - bbox.p.x, bbox.q.y - bbox.p.y,
    gx_no_color_index, color);

  return (code < 0 ? code : 0);
}

private int
vf_release(
  gx_xfont *xf,
  gs_memory_t *mem)
{
  if(mem != NULL)
    gs_free_object(mem, xf, "vflib_release");
  return 0;
}

/* Libraries for vf_draw_font and vf_paint_font */

int
vf_vmatrix(
  int jis_code,
  gs_rect *pbbox,
  gs_matrix *pmat)
{
  return kf_vmatrix(jis_code,
    (floatp)VF_CHAR_CENTER_X, (floatp)(VF_CHAR_CENTER_Y - VF_CHAR_BASE),
    pbbox, pmat);
}

/* following code is derived from zkfvflib.c */
/*
 * zkfvflib.c --- Kanji font operator for VFlib.
 * Copyright (C) ASAYAMA Kazunori  1994.
 */

#include "VF.h"
#ifdef __MSDOS__
#  include "dos_.h"
#endif

#ifndef VFONTCAP_BASE
#  define VFONTCAP_BASE "vfontcap"
#endif
#ifndef MAXPATHLEN
#  define MAXPATHLEN 256
#endif

typedef struct index_item_s {
  struct index_item_s *next;
  char *str;
  int fd;
} index_item;
private index_item *hash_index[256];

/* Hash function */
private int
hash(const byte *str, int len)
{
  unsigned char hash;

  for(hash = 0;len>0; len--,str++)
    hash ^= *str;

  return hash;
}

/* store to hash table */
private int
store(const byte *str, int len, int fd)
{
  int key = hash(str, len);
  index_item *item;

  if((item = (index_item *)malloc(sizeof(index_item))) == 0)
    return -1; /*e_VMerror*/
  if((item->str = (char *)malloc(len+1)) == 0)
    return -1; /*e_VMerror*/
  memcpy(item->str,str,len);
  item->str[len] = 0;
  item->next = hash_index[key];
  item->fd = fd;
  hash_index[key] = item;
  return fd;
}

/* search hash table. */
private int
search(const byte *str, int len)
{
  index_item *item = hash_index[hash(str, len)];

  while(item) {
    if(strlen(item->str) == len && strncmp((const char *)str,item->str,len) == 0)
      return item->fd;		/* Found */
    item = item->next;
  }
  return -1;			/* Not found */
}

/* initialize the library */
private int
init_vflib(void)
{
  char *pfcap;
  int code;

  /* initialize hash table */
  for (code=0;code<256;code++)
    hash_index[code] = NULL;

  pfcap = getenv("VFONTCAP");
  if (pfcap == NULL) { /* environment variable VFONTCAP is not set. */
    char *pfcapb = VFONTCAP_BASE;

    pfcap = (char *)malloc(MAXPATHLEN+3);
    if (pfcap == NULL) return -1; /*e_VMerror*/

    if (!access(pfcapb,0)) { /* found on current directory. */
#ifdef __MSDOS__
      getcwd(pfcap+2,MAXPATHLEN);
      if (strlen(pfcap+2)>=2 && pfcap[3] == ':') {
        /* file name includes drive name. */
        pfcap += 2;
      }
      else {
        union REGS r;
        r.h.ah = 0x19;
        intdos(&r,&r);
        pfcap[0] = 'a' + r.h.al;
        pfcap[1] = ':';
      }
#else
      getcwd(pfcap,MAXPATHLEN);
#endif
      strcat(pfcap,gp_file_name_concat_string(pfcap,strlen(pfcap)));
      strcat(pfcap,pfcapb);
    }
    else {
      FILE *fp;
/*      strcpy(pfcap,pfcapb); */
      fp = lib_ffopen(pfcapb, pfcap, MAXPATHLEN);

      if (fp) fclose(fp);  /* found on GS's searching path. */
      else { /* default vfontcap*/
        free(pfcap);
        pfcap = NULL;
      }
    }
  }
  code = VF_Init(pfcap);	/* Call initializing routine */
  if (code<0) {
    code = -1; /*e_ioerror*/
    eprintf("VFlib: Failure in initialization.\n");
  }
  return code;
}

/*
 * Open font
 */
int
vf_open_font(const byte *font_name, int len, int print_message)
{
  static int initialized_f = 0;
  int code;

  if ( ! initialized_f ) {
    code = init_vflib();
    if (code < 0) return code;	/* error */
    initialized_f = 1; /* Now, initialization done. */
  }

  code = search(font_name, len);
#define FONT_OPEN_FAIL_MARK -1000
  if(code == FONT_OPEN_FAIL_MARK)
    return -1;
  if (code < 0) {
    char *fn = gs_malloc(len+1,1,"VFlib(open_font)");

    if (fn == NULL) return -1; /*e_VMerror*/
    memcpy(fn,font_name, len);
    fn[len] = 0;
    code = VF_OpenFont(fn);
    if (code < 0){
      if(print_message)
        eprintf1("VFlib: Cannot open the font `%s'.\n",fn);
      store(font_name,len,FONT_OPEN_FAIL_MARK);
      code = -1; /*e_ioerror*/
    }
    else {
      code = store(font_name,len,code);
    }
    gs_free(fn,len+1,1,"VFlib(open_font)");
  }

  return code;
}
